;; Xref handling tools
;; see bottom of file for short explanation

;; 2016-01-04: initial release
;; 2016-09-04: XSOURCE and XPATH commands added
;; 2017-09-11: added global settings and lineweight to XDISPLAY

;; ------------------------------------------

;;
;; hide xref and block instances 
;;

(defun C:XHIDE ( / selset n insert bit60) ; xhide_list global
 (if (not (setq selset (ssget "I" '((0 . "INSERT")))))
  (progn
   (princ "\nSelect xref or block instance(s) to hide: ")
   (setq selset (ssget '((0 . "INSERT"))))
  )
 )
 (if selset
  (progn
   (setq n 0)
   (repeat (sslength selset)
    (setq insert (entget (ssname selset n)))
    (if (setq bit60 (assoc 60 insert))
     (setq insert (subst (cons 60 1) bit60 insert))
     (setq insert (append insert (list (cons 60 1))))
    )
    ;(print insert)
    (entmod insert)
    (if (and xhide_list (listp xhide_list))
     (setq xhide_list (append  xhide_list (list (cdr (assoc -1 insert)))))
     (setq xhide_list (list (cdr (assoc -1 insert))))
    )
    (setq n (1+ n))
   )
  )
 )
 (princ)
)


;;
;; unhide xref and block instances
;;

(defun set_xindex (val / )
 (setq n (atoi val))
)

(defun C:XSHOW ( / kwd selset dcl-id n insert bit60 x_list status)
 (setq selset (ssget "X" '((0 . "INSERT")(60 . 1))))
 (if selset
  (progn
   (if (and xhide_list (listp xhide_list))
    (progn
     (initget "All Choose Last")
     (setq kwd (getkword "\nInstances to unhide (All hidden/Last hidden/Choose from list)<Choose>: "))
     (if (not kwd) (setq kwd "Choose"))
    )
    (progn
     (initget "All Choose")
     (setq kwd (getkword "\nInstances to unhide (All hidden/Choose from list)<Choose>: "))
     (if (not kwd) (setq kwd "Choose"))
    )
   )
   (print kwd)
   (cond
    ((= kwd "Last")
     (foreach n xhide_list
      (setq insert (entget (last xhide_list)))
      (if (setq bit60 (assoc 60 insert))
       (setq insert (subst (cons 60 0) bit60 insert))
       (setq insert (append insert (list (cons 60 0))))
      )
      (entmod insert)
     )
     (setq xhide_list (vl-remove (cdr (assoc -1 insert)) xhide_list))
    )
    ((= kwd "All")
     (setq n 0)
     (repeat (sslength selset)
      (setq insert (entget (ssname selset n)))
      (setq insert (subst (cons 60 0) (assoc 60 insert) insert))
      (entmod insert)
      (setq n (1+ n))
     )
    )
    (T
     (setq dcl_id (load_dialog "xrefhandling.dcl"))
     (if (not (new_dialog "ddxshow" dcl_id )) (exit))
     (setq n 0 x_list '())
     (repeat (sslength selset)
      (setq x_list (append x_list (list (cdr (assoc 2 (entget (ssname selset n)))))))
      (setq n (1+ n))
     )
     (setq n nil)
     (start_list "x_list")
     (mapcar 'add_list x_list)
     (end_list)
     (setq status (start_dialog))
     (unload_dialog dcl_id)
     (if (= 1 status)
      (progn
       (if (not n)(setq n 0))
       (setq insert (entget (ssname selset n)))
       (setq insert (subst (cons 60 0) (assoc 60 insert) insert))
       (entmod insert)
      )
     )
    )
   )
  )
  (princ "\nNo hidden instances found. ")
 )
 (princ)
)


;;
;; unload selected xrefs 
;;

(defun C:XUNLOAD ( / selset n insnme argstr) ; xunload_list global
 (if (not (setq selset (ssget "I" '((0 . "INSERT")))))
  (progn
   (princ "\nSelect xref-instance(s) to unload: ")
   (setq selset (ssget '((0 . "INSERT"))))
  )
 )
 (if selset
  (progn
   (setq n 0)
   (repeat (sslength selset)
    (if (eq 4 (logand 4 (cdr (assoc 70 (tblsearch "BLOCK" (setq insnme (cdr (assoc 2 (entget (ssname selset n))))))))))
     (progn
      (if (and xunload_list (listp xunload_list))
       (setq xunload_list (append  xunload_list (list insnme)))
       (setq xunload_list (list insnme))
      )
      (if (or (not argstr) (= "" argstr))
       (setq argstr insnme)
       (setq argstr (strcat argstr "," insnme))
      )
     )
     (princ (strcat "\n" insnme " is not externally referenced"))
    )
    (setq n (+ 1 n))
   )
   (command "_.-XREF" "_UNLOAD" argstr)
   ;(command "_.REGEN")
  )
 )
 (princ)
)


;;
;; reload unloaded xrefs 
;;

(defun C:XRELOAD ( / kwd selset dcl-id n x_name x_data x_list)
 (setq x_list '())
 (tblnext "BLOCK" T)
 (while (setq x_data (tblnext "BLOCK"))
  (if
   (and
    (= 4 (logand 4 (cdr (assoc 70 x_data))))
    (/= 32 (logand 32 (cdr (assoc 70 x_data))))
   )
   (setq x_list (append x_list (list (cdr (assoc 2 x_data)))))
  )
 )
 ;(print x_list)
 (if (= (length x_list) 0)
  (princ "\nNo unloaded references found. ")
  (progn
   (if (and xunload_list (listp xunload_list))
    (progn
     (initget "All Choose Last")
     (setq kwd (getkword "\nInstances to reload (All/Last unloaded/Choose from list)<Choose>: "))
     (if (not kwd) (setq kwd "Choose"))
    )
    (progn
     (initget "All Choose")
     (setq kwd (getkword "\nInstances to reload (All/Choose from list)<Choose>: "))
     (if (not kwd) (setq kwd "Choose"))
    )
   )
   (cond
    ((= kwd "Last")
     (setq x_name (last xunload_list))
     (command "_.-XREF" "_RELOAD" x_name)
     (setq xunload_list (vl-remove x_name xunload_list))
    )
    ((= kwd "All")
     (command "_.-XREF" "_RELOAD" "*")
     (setq xunload_list nil)
    )
    (T
     (setq dcl_id (load_dialog "xrefhandling.dcl"))
     (if (not (new_dialog "ddxreload" dcl_id )) (exit))
     (start_list "x_list")
     (mapcar 'add_list x_list)
     (end_list)
     (setq status (start_dialog))
     (unload_dialog dcl_id)
     (if (= 1 status)
      (progn
       (if (not n)(setq n 0))
       (setq x_name (nth n x_list))
       (command "_.-XREF" "_RELOAD" x_name)
       (setq xunload_list (vl-remove x_name xunload_list))
      )
     )
    )
   )
  )
 )
 (princ)
)


;;
;; reload (update) loaded xrefs 
;;

(defun C:XUPDATE ( / selset n insnme x_list argstr)
 (if (not (setq selset (ssget "I" '((0 . "INSERT")))))
  (progn
   (princ "\nSelect xref-instance(s) to reload (return for all): ")
   (setq selset (ssget '((0 . "INSERT"))))
  )
 )
 (if selset
  (progn
   (setq n 0)
   (setq x_list "")
   (repeat (sslength selset)
    (if (eq 4 (logand 4 (cdr (assoc 70 (tblsearch "BLOCK" (setq insnme (cdr (assoc 2 (entget (ssname selset n))))))))))
     (setq x_list (strcat x_list "," insnme))
     (princ (strcat "\n" insnme " is not externally referenced"))
    )
    (setq n (+ 1 n))
   )
   (if (/= "" x_list)
    (command "_.-XREF" "_RELOAD" x_list)
   )
  )
  (progn
   (setq x_list '())
   (tblnext "BLOCK" T)
   (while (setq x_data (tblnext "BLOCK"))
    (if
     (and
      (= 4 (logand 4 (cdr (assoc 70 x_data))))
      (= 32 (logand 32 (cdr (assoc 70 x_data))))
     )
     (setq x_list (append x_list (list (cdr (assoc 2 x_data)))))
    )
   )
   (print x_list)
   (if (= (length x_list) 0)
    (princ "\nNo loaded external references found. ")
    (progn
     (foreach n x_list
      (if (or (not argstr) (= "" argstr))
      (setq argstr n)
      (setq argstr (strcat argstr "," n))
      )
     )
     ;(print argstr)
     (command "_.-XREF" "_RELOAD" argstr)
    )
   )
  )
 )
 (princ)
)


;;
;; change xref display
;;

(defun xdisplay-settings ( / dcl_id xretain xmode xfade xfdlst status)
 (setq dcl_id (load_dialog "xrefhandling.dcl"))
 (if (not (new_dialog "ddxsettings" dcl_id )) (exit))
 (setq xretain (getvar "VISRETAIN"))
 (cond
  ((= 1 xretain) (set_tile "xretain" "1"))
  ((= 0 xretain) (set_tile "xretain" "0"))
  (t (mode_tile "xretain" 1))
 )
 (setq xmode (getvar "XREFOVERRIDE"))
 (cond
  ((= 1 xmode) (set_tile "xmode" "1"))
  ((= 0 xmode) (set_tile "xmode" "0"))
  (t (mode_tile "xmode" 1))
 )
 (setq xfdlst '(" " "0" "10" "20" "30" "40" "50" "60" "70" "80" "90"))
 (start_list "xfadelist")
 (mapcar 'add_list xfdlst)
 (end_list)
 (if (getvar "XDWGFADECTL")
  (set_tile "xfadelist" (itoa (vle-search (itoa (* (/ (getvar "XDWGFADECTL") 10) 10)) xfdlst T)))
 )
 (setq status (start_dialog))
 (unload_dialog dcl_id)
 (if (= 1 status)
  (progn
   (if xmode (setvar "XREFOVERRIDE" xmode))
   (if xretain (setvar "VISRETAIN" xretain))
   (if xfade (setvar "XDWGFADECTL" xfade))
   (command "_.REGEN")
  )
 )
)


(defun set_matname (val / )
 (setq matnme (nth (atoi val) matlst))
)

(defun set_lcolor ( / )
 ;(setq lcolor (acad_truecolordlg '(62 . 7) nil))
 (if (setq lcolor (acad_colordlg 7 T))
  (progn
   (start_image "color_button")
   (if (or (= lcolor 0) (= lcolor 256))
    (setq lcolor -15)
   )
   (fill_image
    0
    0
    (dimx_tile "color_button")
    (dimy_tile "color_button")
    lcolor
   )
   (end_image)
   (mode_tile "color_button" 1) ; hack to properly repaint the button (on linux)
   (mode_tile "color_button" 0)
  )
 )
)

(defun set_llw (val / )
 (setq llw (nth (atoi val) lwlst))
 (if (< 1 (atoi val))
  (setq llw (substr llw 1 4))
 )
)

(defun set_ltransparency (val / )
 (setq ltrans (nth (atoi val) trnlst))
)

(defun set_xfade (val / )
 (setq xfade (atoi (nth (atoi val) xfdlst)))
)

(defun set_xretain (val / )
 (setq xretain (atoi val))
)

(defun set_xmode (val / )
 (setq xmode (atoi val))
)

(defun C:XDISPLAY ( / selset n insnme x_list dcl_id matlst matnme matdct lcolor trnlst lwlst llw ltrans status)
 (if (not (setq selset (ssget "I" '((0 . "INSERT")))))
  (progn
   (princ "\nSelect xref-instance(s) for layer property overrides: (Return for global settings)")
   (setq selset (ssget '((0 . "INSERT"))))
  )
 )
 (if selset
  (progn
   (setq n 0)
   (repeat (sslength selset)
    (if (eq 4 (logand 4 (cdr (assoc 70 (tblsearch "BLOCK" (setq insnme (cdr (assoc 2 (entget (ssname selset n))))))))))
     (setq x_list (cons insnme x_list))
     (princ (strcat "\n" insnme " is not externally referenced"))
    )
    (setq n (+ 1 n))
   )
   (if x_list
    (progn
     (setq n nil)
     (setq dcl_id (load_dialog "xrefhandling.dcl"))
     (if (not (new_dialog "ddxdisplay" dcl_id )) (exit))
     (setq exclst (list "_DYN_DIM_3D" "BmFeatureColorMaterial" "ByBlock" "ByLayer"))
     (setq matdct (dictsearch (namedobjdict) "ACAD_MATERIAL"))
     (foreach n (dictsearch (namedobjdict) "ACAD_MATERIAL")
      (if (and (= (car n) 3)(not (member (cdr n) exclst)))
       (setq matlst (cons (cdr n) matlst))
      )
     )
     (setq matlst (cons " " matlst))
     (start_list "matlist")
     (mapcar 'add_list matlst)
     (end_list)
     (setq trnlst '(" " "0" "10" "20" "30" "40" "50" "60" "70" "80" "90"))
     (start_list "translist")
     (mapcar 'add_list trnlst)
     (end_list)
     (setq lwlst '(" " "default" "0.00 mm" "0.10 mm" "0.20 mm" "0.30 mm" "0.40 mm" "0.50 mm" "0.60 mm" "0.70 mm" "0.80 mm" "0.90 mm" "1.00 mm" "1.20 mm" "1.40 mm" "1.60 mm" "2.00 mm"))
     (start_list "lwlist")
     (mapcar 'add_list lwlst)
     (end_list)
     (setq status (start_dialog))
     (unload_dialog dcl_id)
     (if (= 1 status)
      (progn
       (command "_.-LAYER")
       (foreach n x_list
        (if (and matnme (/= " " matnme))
         (command "_MATERIAL" matnme (strcat n "|*"))
        )
        (if (and lcolor (< 0 lcolor 256))
         (command "_COLOR" lcolor (strcat n "|*"))
        )
        (if (and llw (/= " " llw))
         (if (= 0 (getvar "LWUNITS"))
          (command "_LW" (rtos (/ (atof llw) 25.4)) (strcat n "|*"))
          (command "_LW" llw (strcat n "|*"))
         )
        )
        (if (and ltrans (/= " " ltrans))
         (command "_TRANSPARENCY" ltrans (strcat n "|*"))
        )
       )
       (command "")
       (command "_.REGEN")
      )
     )
    )
    (princ "\nNo externally referenced inserts found. ")
   )
  )
  (xdisplay-settings)
 )
 (princ)
)


;;
;; reset selected xrefs to their original layer settings
;;

(defun C:XRESET ( / selset n insnme xreset_list visretain)
 (if (not (setq selset (ssget "I" '((0 . "INSERT")))))
  (progn
   (princ "\nSelect xref-instance(s) to reload with original layer settings (return for all): ")
   (setq selset (ssget '((0 . "INSERT"))))
  )
 )
 (if selset
  (progn
   (setq n 0)
   (setq xreset_list "")
   (repeat (sslength selset)
    (if (eq 4 (logand 4 (cdr (assoc 70 (tblsearch "BLOCK" (setq insnme (cdr (assoc 2 (entget (ssname selset n))))))))))
     (setq xreset_list (strcat xreset_list "," insnme))
     (princ (strcat "\n" insnme " is not externally referenced"))
    )
    (setq n (+ 1 n))
   )
   (if (/= "" xreset_list)
    (progn
     (setq visretain (getvar "VISRETAIN"))
     (setvar "VISRETAIN" 0)
     (command "_.-XREF" "_RELOAD" xreset_list)
     (setvar "VISRETAIN" visretain)
    )
   )
  )
  (progn
   (setq visretain (getvar "VISRETAIN"))
   (setvar "VISRETAIN" 0)
   (command "_.-XREF" "_RELOAD" "*")
   (setvar "VISRETAIN" visretain)
  )
 )
 (princ)
)

;;
;; xref path conversion
;;

(defun xpath_full_2_relative (dwgdir xpath / ddrive xdrive p q dlst plst rpath slash n)
 (setq
      dlst (vle-string-split ":" dwgdir)
      plst (vle-string-split ":" xpath)
 )
 (if (< 1 (length dlst)) ; path contains drive letter
  (setq ddrive (car dlst) dwgdir (cadr dlst))
 )
 (if (< 1 (length plst))
  (setq xdrive (car plst) plst (cadr plst))
 )
 (if (not (eq ddrive xdrive)) ; different drive letters or something went wrong, return full path
  xpath
  (progn
   (setq
	 dlst (vle-remove-all "" (vle-string-split "/\\" dwgdir))
	 plst (vle-remove-all "" (vle-string-split "/\\" (car plst)))
	 rpath ""
	 slash "/" ; should perhaps be set to "\\" on MS-platforms
   )
   (while (eq (car dlst) (car plst)) ; should we rather do a case-insensitive compare?
    (setq dlst (cdr dlst) plst (cdr plst))
   )
   (foreach n plst
    (setq rpath
     (if (= "" rpath)
      n
      (strcat rpath slash n)
     )
    )
   )
   (setq n (length dlst))
   (if (= n 0)
    (setq rpath (strcat "." slash rpath))
    (repeat n (setq rpath (strcat ".." slash rpath)))
   )
  )
 )
)


(defun xpath_relative_2_full (dwgdir rpath / dlst plst xpath slash smb n)
 (if (vl-string-position (ascii "\\") dwgdir) ; looks we are on a MS box
  (setq
        slash "\\"
	smb (substr dwgdir 1 2)
  )
  (setq slash "/")
 )
 (setq
       dlst (reverse (vle-remove-all "" (vle-string-split "/\\" dwgdir)))
       plst (vle-remove-all "" (vle-string-split "/\\" rpath))
       xpath ""
 )
 (if (eq (car plst) ".")
  (progn
   (setq plst (cdr plst) xpath (vl-string-right-trim "/\\" dwgdir))
   (foreach n plst
    (setq xpath (strcat xpath slash n))
   )
  )
  (progn
   (while (eq (car plst) "..")
    (setq dlst (cdr dlst) plst (cdr plst))
   )
   (setq plst (append (reverse dlst) plst))
   (foreach n plst
    (setq xpath
     (cond
      ((and (= "" xpath)(= "\\\\" smb))
       (strcat smb n)
      )
      ((and (= "" xpath)(= "\\" slash))
      n
      )
      (t
       (strcat xpath slash n)
      )
     )
    )
   )
  )
 )
)


(defun C:XPATH ( / selset n xpath dwgdir ptype insnme)
 (if (not (setq selset (ssget "I" '((0 . "INSERT")))))
  (progn
   (princ "\nSelect xref-instance(s) for path conversion: ")
   (setq selset (ssget '((0 . "INSERT"))))
  )
 )
 (if selset
  (progn
   (setq n 0)
   (setq dwgdir (getvar "dwgprefix"))
   (initget "Absolute Relative None")
   (setq ptype (getkword "\nConvert xref-path to (Absolute / Relative / None) <Relative>: "))
   (repeat (sslength selset)
    (if (setq xpath (cdr (assoc 1 (tblsearch "BLOCK" (setq insnme (cdr (assoc 2 (entget (ssname selset n)))))))))
     (progn
      (cond
       ((= "Absolute" ptype)
        (cond
	 ((eq (substr xpath 1 1) ".")
	  (setq xpath (xpath_relative_2_full dwgdir xpath))
	 )
	 ((not (or (eq (substr xpath 1 1) "/") (eq (substr xpath 1 1) "\\") (wcmatch xpath "@:*"))) ; no path
	  (setq xpath (findfile xpath))
	 )
	 (t (setq xpath nil))
	)
       ) 
       ((= "None" ptype)
        (cond
	 ((or (eq (substr xpath 1 1) "/") (eq (substr xpath 1 1) "\\") (wcmatch xpath "@:*"))
	  (if (findfile (acet-filename-path-remove xpath))
	   (setq xpath (acet-filename-path-remove xpath))
	   (setq xpath (xpath_full_2_relative dwgdir xpath))
	  )
	 )
	 ((eq (substr xpath 1 1) ".")
	  (if (findfile (acet-filename-path-remove xpath))
	   (setq xpath (acet-filename-path-remove xpath))
	   (setq xpath nil)
	  )
	 )
	 (t (setq xpath nil))
	)
       )
       (t ; relative path
	(cond
	 ((or (eq (substr xpath 1 1) "/") (eq (substr xpath 1 1) "\\") (wcmatch xpath "@:*"))
	  (setq xpath (xpath_full_2_relative dwgdir xpath))
	 )
	 ((/= (substr xpath 1 1) ".")
	  (if (setq xpath (findfile xpath))
	   (setq xpath (xpath_full_2_relative dwgdir xpath))
	   (setq xpath nil)
	  )
	 )
	 (t (setq xpath nil))
	)
       )
      )
      (if (and xpath (findfile xpath))
       (command "_.-XREF" "_PATH" insnme xpath)
      )
     )
     ;(princ (strcat "\nBlock " insnme " is not externally referenced."))
    )
    (setq n (1+ n))
   )
  )
  (princ "\nNo externally referenced inserts found.")
 )
 (princ)
)

;;
;; select xref source file
;;

(defun C:XSOURCE ( / selset xpath dwgdir insnme)
 (if (or (not (setq selset (ssget "I" '((0 . "INSERT"))))) (< 1 (sslength selset)))
  (progn
   (princ "\nSelect xref-instance for source selection: ")
   (setq selset (ssget ":S" '((0 . "INSERT"))))
  )
 )
 (if selset
  (if (setq xpath (cdr (assoc 1 (tblsearch "BLOCK" (setq insnme (cdr (assoc 2 (entget (ssname selset 0)))))))))
   (progn
    (print xpath)
    (setq dwgdir (getvar "dwgprefix"))
    (cond
     ((eq (substr xpath 1 1) ".") ; relative path
      (setq xpath (xpath_full_2_relative dwgdir (getfiled (strcat "Select source for " insnme) (xpath_relative_2_full dwgdir xpath) "dwg" 0)))
     ) 
     ((or (eq (substr xpath 1 1) "/") (eq (substr xpath 1 1) "\\") (wcmatch xpath "@:*")) ; absolute path
      (setq xpath (getfiled (strcat "Select source for " insnme) xpath "dwg" 0))
     )
     (t ; no path
      (if (setq xpath (findfile xpath))
       (setq xpath (getfiled (strcat "Select source for " insnme) xpath "dwg" 0))
       (setq xpath (getfiled (strcat "Select source for " insnme) dwgdir "dwg" 0))
      )
      (print xpath)
      (if (eq (findfile (acet-filename-path-remove xpath)) xpath)
       (setq xpath (acet-filename-path-remove xpath))
       (setq xpath (xpath_full_2_relative dwgdir xpath))
      )
     )
    )
    (command "_.-XREF" "_PATH" insnme xpath)
    (command "_.REGEN")
   )
   (princ "\nNo externally referenced inserts found. ")
  )
 )
 (princ)
)




;; ------------------------------------------

(prompt
 "\n\nXref handling routines
  ----------------------------------
  XHIDE: hides block/xref instances
  XSHOW: unhides block/xref instances
  XUNLOAD: unloads external references
  XRELOAD: reloads unloaded external references
  XUPDATE: reloads loaded external references
  XDISPLAY: overrides xref layer properties
  XRESET: resets xref layer overrides
  XSOURCE: selects xref source file
  XPATH: converts between absolute/relative/no xref paths 
  ----------------------------------
  2016-2017 knut hohenberg www.khad.eu
  no rights reserved, no responsability taken!
 "
 )
 (princ)
